/*
 * Tree.java
 *
 * Created on November 13, 2003, 8:33 AM
 */

package edu.umiacs.fmm;
import java.util.*;
/**
 *
 * @author  wpwy
 */
public class FmmTree{
    public static final int MAX_NUM_LEVEL=8;
    final int dimension = 2;
    //int numOfLevels;
    Vector<Point> x;
    Vector<Point> y;
    Potential potential;
    Vector<Vector<Box>> struct;
    Box root;
    int q;
    
    public static final int DEFAULT_NUM_LEVEL = 3;
    int currLevel;
    
    long numTrans;
    long numExp;
    long numDirect;
    
    /** Creates a new instance of Tree */
    public FmmTree() {
        //numOfLevels = DEFAULT_NUM_LEVEL;
        currLevel = DEFAULT_NUM_LEVEL-1;
        x = new Vector<Point>();
        y = new Vector<Point>();
        struct = new Vector<Vector<Box>>();
        potential = new Potential();
    }
    public FmmTree(Vector<Point> x, Vector<Point> y, Potential p){
        //default level of division
        this.x = x;
        this.y = y;
        this.potential = p;
        struct = new Vector<Vector<Box>>();
        //numOfLevels = DEFAULT_NUM_LEVEL;
        initDefaultStruct();
        
    }
    private void initDefaultStruct(){
        int p = potential.getP();
        root = new Box(0, 0, p);
        Vector<Box> layer0 = new Vector<Box>(1);
        layer0.addElement(root);
        struct.addElement(layer0);
        
        Vector<Box> layer1 = new Vector<Box>(4);
        for (int i=0; i<4; i++){
            Box child = new Box(1, i, p);
            child.setParent(root);
            root.addChild(child);
            layer1.addElement(child);
        }
        struct.addElement(layer1);
        
        Vector<Box> layer2 = new Vector<Box>(16);
        for (int i=0; i<16; i++){
            Box child = new Box(2, i, p);
            Box parent = getBox(1, i>>2);
            child.setParent(parent);
            parent.addChild(child);
            layer2.addElement(child);
        }
        struct.addElement(layer2);
        
        //---tmp test start
        /*
        Vector<Box> layer3 = new Vector<Box>(64);
        for (int i=0; i<64; i++){
            Box child = new Box(3, i, p);
            Box parent = getBox(2, i>>2);
            child.setParent(parent);
            parent.addChild(child);
            layer3.addElement(child);
        }
        struct.addElement(layer3);
         */
        //---tmp test end
        
        //now set leaf data;
        int lastLevelIndex = struct.size()-1;
        for (Point px:x)
            struct.lastElement().elementAt(px.getBoxIndex(lastLevelIndex)).addX(px);
        for (Point py:y)
            struct.lastElement().elementAt(py.getBoxIndex(lastLevelIndex)).addY(py);
    }
    /*
     public Box getParent(FmmTree t) {
        return t.getBox(level-1, index >> 2);
    }
    public Box[] getChildren(FmmTree t) {
        try {
            Box[] ans = new Box[4];
            for (int i=0; i<4; i++){
                ans[i] = t.getBox(level+1, (index<<2)+i);
            }
            return ans;
     
        }
        catch (Exception e){
            return null;
        }
    }
     */
    
    public static FmmTree build(int th, Point[] x, Point[] y, Potential p){
        Vector<Point> xv = new Vector<Point>();
        Vector<Point> yv = new Vector<Point>();
        for (Point px:x)
            xv.addElement(px);
        for (Point py:y)
            yv.addElement(py);
        return build(th, xv, yv, p);
    }
    public static FmmTree build(int maxClusterThreshold, Vector<Point> x, Vector<Point> y, Potential potential) {
        if (maxClusterThreshold<=0){
            return build(x, y, potential);
        }
        FmmTree ans = new FmmTree(x, y, potential);
        //System.out.println("After default init, tree is now: ");
        //System.out.println(ans.treeToString());
        for (Box b:ans.struct.lastElement()){
            partition(ans.struct, b, maxClusterThreshold);
        }
        ans.buildL1();
        ans.buildL2L3L4();
        //System.out.println(ans.treeToString());
        //System.out.println(ans.dTreeToString());
        ans.setQ(maxClusterThreshold);
        return ans;
    }
    
    public void buildL1(){
        for (int i=struct.size()-1; i>0; i--){
            Vector<Box> layer = struct.elementAt(i);
            for (Box b:layer){
                if (!b.hasChildren() && !b.isDummy()){//is a leaf node
                    Vector<Box> e2 = b.getNeighborsAllNonDummy(this);
                    for (Box nb:e2){
                        if (!nb.hasChildren()){
                            nb.addL1(b);
                            b.addL1(nb);
                        }
                    }
                } else if (b.hasChildren()){
                    //build L1b
                    Vector<Box> e2 = b.getNeighborsAllNonDummy(this);
                    for (Box nb:e2){
                        if (!nb.hasChildren()){
                            nb.addL1b(b);
                            b.addL1b(nb);
                        }
                    }
                }
            }
        }
    }
    public void buildL2L3L4(){
        for (int i=struct.size()-1; i>0; i--){
            Vector<Box> layer = struct.elementAt(i);
            for (Box b:layer){
                if (!b.isDummy()){
                    
                    Vector<Box> e4 = b.getNeighborsE4IncDummy(this);
                    for (Box c:e4){
                        if (!c.isDummy()){
                            b.addL2(c);
                            continue;
                        }
                        while (c.isDummy()){
                            c=getBox(c.getLevel()-1, c.getIndex()>>2);
                        }
                        if (  (!b.hasChildren() && !b.getL1().contains(c)) ||
                                b.hasChildren() && !b.getL1b().contains(c)) {
                            c.addL3(b);
                            b.addL4(c);
                        }
                        /*
                        if (e4b.isDummy()){
                            //System.out.println(e4b+" is dummy");
                            while (e4b.isDummy()){
                                int parentLevel = e4b.getLevel()-1;
                                int parentIndex = e4b.getIndex()>>2;
                                e4b = getBox(parentLevel, parentIndex);
                            }
                            if (!b.getL1().contains(e4b)){
                                e4b.addL3(b);
                                b.addL4(e4b);
                            }
                        } else  {
                            b.addL2(e4b);
                        }
                         */
                    }
                }
            }
        }
    }
    
    private static void partition(Vector<Vector<Box>> struct, Box b, int q){
        //System.out.println("Trying to partition "+b.toLongString());
        int xCount = b.getX().size();
        //int yCount = b.getY().size();
        if (xCount>q){
            //need to split
            //System.out.println("Need to split");
            int childLevel = b.getLevel()+1;
            //if new level does not already exist, create a new level and
            //fill it with dummies.
            
            if (struct.size()==(b.getLevel()+1)){//need to add a new level
                int size = (int)(Math.pow(4, struct.size()));
                Vector<Box> layer = new Vector<Box>(size);
                for (int i=0; i<size; i++) {
                    Box dummyBox = new Box(childLevel, i, b.getP());
                    dummyBox.setDummy(true);
                    //----tmp fix
                    //Box dummyParent = struct.elementAt(childLevel-1).elementAt(i >> 2);
                    //dummyBox.setParent(dummyParent);
                    //dummyParent.addChild(dummyBox);
                    //---end tmp fix
                    layer.addElement(dummyBox);
                }
                struct.addElement(layer);
            }
            //now split
            Vector<Box> children = new Vector<Box>(4);
            Vector<Point> xv = b.clearX();
            Vector<Point> yv = b.clearY();
            for (int i=0; i<4; i++){
                Box child = new Box(childLevel, (b.getIndex()<<2)+i, b.getP());
                child.setParent(b);
                b.addChild(child);
                struct.elementAt(childLevel).setElementAt(child, child.getIndex());
                children.addElement(child);
            }
            //now put in the x y points;
            
            for (Point px:xv){
                struct.elementAt(childLevel).elementAt(px.getBoxIndex(childLevel)).addX(px);
            }
            for (Point py:yv){
                struct.elementAt(childLevel).elementAt(py.getBoxIndex(childLevel)).addY(py);
            }
            //now recursion
            for (Box child:children){
                if (child.getLevel()<FmmTree.MAX_NUM_LEVEL)
                    partition(struct, child, q);
            }
        }
        //otherwise do nothing
    }
    
    public static FmmTree build(Point[] x, Point[] y, Potential p){
        Vector<Point> xv = new Vector<Point>();
        Vector<Point> yv = new Vector<Point>();
        for (Point px:x)
            xv.addElement(px);
        for (Point py:y)
            yv.addElement(py);
        return build(xv, yv, p);
    }
    public static FmmTree build(Vector<Point> x, Vector<Point> y, Potential p){
        int N = x.size();
        int dl = Math.round((float)(Math.log(N)/Math.log(2)/2));
        int q = Math.round((float)(9*N*Math.pow(2, -dl)));
        //System.out.println("Clustering: "+q);
        return build(q, x, y, p);
    }
    public void setQ(int q){
        this.q=q;
    }
    public int getQ(){
        return q;
    }
    
    
    private void printComplex(Complex[] c){
        for (Complex cc:c)
            System.out.print(cc+" ");
    }
    
    private void upwardPass(double[] u){
        for (int l=struct.size()-1; l>1; l--){
            for (Box b:struct.elementAt(l)){
                if (!b.isDummy() && b.getXRecursive().size()>0){
                    if (!b.hasChildren()){ //leaf node
                        for(Point px:b.getX()){
                            this.numExp++;
                            Complex[] B = potential.getSCoeff(px.getCoord(), b.getCenter().getCoord());
                            double thisU = u[x.indexOf(px)];
                            for (int k=0; k<B.length; k++) {
                                B[k]=B[k].multiply(thisU);
                            }
                            b.addC(B);
                        }
                    } else {//internal node
                        for (Box childB:b.getChildren()) {
                            if (childB.getXRecursive().size()>0){
                                Complex from = childB.getCenter().getCoord();
                                Complex to = b.getCenter().getCoord();
                                b.addC(potential.getSS(from, to, childB.getC()));
                                this.numTrans++;
                            }
                        }
                    }
                }
            }
        }
    }
    private void downwardPass(double[] u){
        for (int l=2; l<struct.size(); l++){
            for (Box b:struct.elementAt(l)){
                //System.out.println("Looking at Box "+b);
                if (!b.isDummy() && b.getYRecursive().size()>0){
                    
                    Complex bCenter = b.getCenter().getCoord();
                    for (Box l2b:b.getL2()){
                        if (l2b.getXRecursive().size()>0){
                            Complex from = l2b.getCenter().getCoord();
                            b.addD(potential.getSR(from, bCenter, l2b.getC()));
                            numTrans++;
                        }
                    }
                    for (Box l4b:b.getL4()){
                        for (Point px:l4b.getX()){
                            Complex[] B = potential.getRCoeff(px.getCoord(), bCenter);
                            double thisU = u[x.indexOf(px)];
                            for (int k=0; k<B.length; k++) {
                                B[k]=B[k].multiply(thisU);
                            }
                            b.addD(B);
                            numExp++;
                        }
                    }
                    if (b.hasChildren()){
                        for (Box child:b.getChildren()){
                            if (child.getYRecursive().size()>0){
                                Complex to = child.getCenter().getCoord();
                                child.addD( potential.getRR(bCenter, to, b.getD()));
                                numTrans++;
                            }
                        }
                    }
                }
            }
        }
    }
    
    public double[] solve(double[] u) {
        upwardPass(u);
        downwardPass(u);
        double[] v = new double[y.size()];
        solveHelper(u, v);
        
        return v;
    }
    
    private void solveHelper(double[] u, double[] v){
        for (int i=struct.size()-1; i>0; i--){
            for (Box b:struct.elementAt(i)){
                if (!b.isDummy()){
                    if (!b.hasChildren()){ //leaf
                        Complex[] d = b.getD();
                        
                        Complex bCenter = b.getCenter().getCoord();
                        for (Point py:b.getY()){
                            int idx = y.indexOf(py);
                            Complex[] rVec = potential.getRVector(py.getCoord(), bCenter);
                            numExp++;
                            for (int k=0; k<d.length; k++){
                                v[idx] += d[k].multiply(rVec[k]).real();
                            }
                            Vector<Box> l1bAndb = new Vector<Box>();
                            l1bAndb.addAll(b.getL1());
                            l1bAndb.add(b);
                            for (Box l1b:l1bAndb){
                                for (Point px:l1b.getX()){
                                    double thisU = u[x.indexOf(px)];
                                    double direct = potential.direct(py.getCoord(), px.getCoord()).multiply(thisU).real();
                                    v[idx]+=direct;
                                    numDirect++;
                                }
                            }
                            //System.out.println("summing up L3");
                            for (Box l3b:b.getL3()){
                                //System.out.println("L3 box found");
                                if (l3b.getXRecursive().size()>0){
                                    Complex[] sVec = potential.getSVector(py.getCoord(), l3b.getCenter().getCoord());
                                    Complex[] c = l3b.getC();
                                    numExp++;
                                    for (int k=0; k<c.length; k++){
                                        v[idx] += c[k].multiply(sVec[k]).real();
                                    }
                                }
                            }
                            //System.out.println("Done summing up L3");
                        }
                    }
                }
            }
        }
    }
    
    /*
    private void solveHelper(Box thisBox, double[] v, double[] u){
        //System.out.println("solving: "+thisBox);
        if (thisBox.isDTreeLeaf()){
            //System.out.println("is leave");
            for (Point thisY:thisBox.getYRecursive()){
                double regPart = 0d;
                Complex[] d = thisBox.getD();
                Complex[] rVec = potential.getRVector(thisY.getCoord(), thisBox.getCenter().getCoord());
                numTrans++;
                for (int k=0; k<d.length; k++){
                    regPart += d[k].multiply(rVec[k]).real();
                }
     
                double sinPart = 0d;
                Vector<Box> nei = thisBox.getNeighbors(this);
                Vector<Box> neiandme = new Vector<Box>(nei);
                neiandme.addElement(thisBox);
     
                for (Box neibox:neiandme){
                    for (Point thisX:  neibox.getXRecursive()) {
                        double thisU = u[x.indexOf(thisX)];
                        double inc = potential.direct(thisY.getCoord(), thisX.getCoord()).multiply(thisU).real();
                        sinPart+=inc;
                    }
                }
                v[y.indexOf(thisY)] = sinPart+regPart;
                //System.out.println("Setting v["+y.indexOf(thisY)+"] to "+(sinPart+regPart));
            }
        }
     
        else {
            //System.out.println("not leave");
            for(Box child:thisBox.getDTreeChildren()) {
                solveHelper(child, v, u);
            }
        }
    }
     
     */
    
    public double[] solveDirect(double[] u) {
        double[] v = new double[y.size()];
        
        for (int j=0; j<v.length; j++) {
            for (int i=0; i<x.size(); i++) {
                v[j]+=u[i]* potential.direct(y.elementAt(j).getCoord(), x.elementAt(i).getCoord()).real();
                //numOpsDirect++;
            }
        }
        
        return v;
    }
    
    
    //public int getClusterThreshold() {
    //    int ans = 0;
        /*for (int i=0; i<struct[numOfLevels-1].length; i++) {
            int xlength = struct[numOfLevels-1][i].getX().length;
            int ylength = struct[numOfLevels-1][i].getY().length;
            if (xlength>ans)
                ans = xlength;
            if (ylength>ans)
                ans = ylength;
        }
         */
    //    return ans;
    //}
    
    public Box getRoot(){
        return struct.firstElement().firstElement();
    }
    public Vector<Vector<Box>> getStruct(){
        return struct;
    }
    
    public Box getBox(int level, int index) {
        try {
            return struct.elementAt(level).elementAt(index);
        } catch (Exception e){
            return null;
        }
    }
    public int getNumOfLevels(){
        return struct.size();
    }
    public void print(){
        int l=0;
        for (Vector<Box> level : struct){
            System.out.println("########Level "+(l++)+": ");
            for (Box b:level) {
                if (!b.isDummy()){
                    System.out.println(b.toString());
                    System.out.print("    L1: ");
                    for (Box bb:b.getL1()){
                        System.out.print(bb.toString()+" ");
                    }
                    System.out.println();
                    
                    System.out.print("    L2: ");
                    for (Box bb:b.getL2()){
                        System.out.print(bb.toString()+" ");
                    }
                    System.out.println();
                    
                    System.out.print("    L3: ");
                    for (Box bb:b.getL3()){
                        System.out.print(bb.toString()+" ");
                    }
                    System.out.println();
                    
                    System.out.print("    L4: ");
                    for (Box bb:b.getL4()){
                        System.out.print(bb.toString()+" ");
                    }
                    System.out.println();
                    System.out.print("    X points: ");
                    for (Point xp:b.getX())
                        System.out.print(xp.toString()+" ");
                    System.out.println();
                    System.out.print("    Y points: ");
                    for (Point yp:b.getY())
                        System.out.print(yp.toString()+" ");
                    System.out.println();
                    
                    System.out.print("    C values: ");
                    for (Complex c:b.getC()) {
                        System.out.print(c.toString()+" ");
                    }
                    System.out.println();
                    System.out.print("    Dtilde values: ");
                    for (Complex c:b.getDtilde()) {
                        System.out.print(c.toString()+" ");
                    }
                    System.out.println();
                    System.out.print("    D values: ");
                    for (Complex c:b.getD()) {
                        System.out.print(c.toString()+" ");
                    }
                    System.out.println();
                }
            }
            System.out.println("########End Level "+l+"\n");
        }
    }
    public String toString() {
        String ans = "";
        int l=0;
        for (Vector<Box> level : struct){
            ans+="########Level "+(l++)+": "+"\n";
            for (Box b:level) {
                if (b!=null){
                    ans+=b.toString()+"\n";
                    ans+="X points: ";
                    for (Point xp:b.getX())
                        ans+=xp.toString()+" ";
                    ans+="\n";
                    ans+="Y points: ";
                    for (Point yp:b.getY())
                        ans+=yp.toString()+" ";
                    ans+="\n";
                    
                    ans+="C values: ";
                    for (Complex c:b.getC()) {
                        ans+=c.toString()+" ";
                    }
                    ans+="\n";
                    ans+="Dtilde values: ";
                    for (Complex c:b.getDtilde()) {
                        ans+=c.toString()+" ";
                    }
                    ans+="\n";
                    ans+="D values: ";
                    for (Complex c:b.getD()) {
                        ans+=c.toString()+" ";
                    }
                    ans+="\n";
                } else {
                    ans +="NULL\n";
                }
            }
            ans+="########End Level "+l+"\n\n";
        }
        return ans;
    }
    
    
    public String treeToString(){
        return treeToStringHelper(getRoot(), "  ");
    }
    
    private String treeToStringHelper(Box root, String prefix){
        //System.out.println("looking at "+root.toString());
        if (root.getChildren().size()==0)
            return prefix + root.toLongString()+"\n";
        else {
            String ans = prefix+root.toLongString()+"\n";
            
            ans+=prefix+"begin "+root.toString()+" children\n";
            for (Box b:root.getChildren())
                ans+=treeToStringHelper(b, prefix+prefix);
            ans+=prefix+"end "+root.toString()+" children\n";
            return ans;
        }
    }
    
    public Vector<Point> getX() {
        return x;
    }
    public Vector<Point> getY(){
        return y;
    }
    public long getNumTrans(){
        return numTrans;
    }
    public void setNumTrans(long i){
        this.numTrans=i;
    }
    public long getNumExp(){
        return numExp;
    }
    public void setNumExp(long i){
        this.numExp=i;
    }
    public long getNumDirect(){
        return numDirect;
    }
    public void setNumDirect(long i){
        this.numDirect=i;
    }
    public Potential getPotential(){
        return potential;
    }
    public void setPotential(Potential p){
        this.potential = p;
    }
}

