/*
 * Main.java
 *
 * Created on November 9, 2003, 12:45 PM
 */

package edu.umiacs.fmm;
import java.util.*;
import java.io.*;
/**
 *
 * @author  wpwy
 */
public class Main {
    
    /** Creates a new instance of Main */
    public Main() {
    }
    
    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception {
        int N=500;
        int threshold = 4;
        int p=5;
        
        Vector<Point> x=new Vector<Point>(N);
        Vector<Point> y=new Vector<Point>(N);
        double[] u = new double[N];
        
        /*
        for (int i=0; i<N; i++) {
            Point pt1 = new Point(Math.random(), Math.random());
            Point pt2 = new Point(Math.random(), Math.random());
            x.addElement(pt1);
            y.addElement(pt2);
            u[i] = Math.random();
        }
         
         */
         
        loadInput(x, y, u);
        
        FmmTree t = FmmTree.build(threshold, x, y, new Potential(p));
        FmmTreeNonAdaptive tt = FmmTreeNonAdaptive.build(t.getNumOfLevels(), x.toArray(new Point[0]), y.toArray(new Point[0]), new Potential(p));
        double[] indirect = t.solve(u);
        double[] indirectNA = tt.solve(u);
        
        
        //FmmTreeNonAdaptive tt = FmmTreeNonAdaptive.build(threshold, x, y, new Potential(p));
        double[] direct = t.solveDirect(u);
        double error = 0;
        for (int i=0; i<direct.length; i++) {
            //System.out.println(direct[i]+" "+indirect[i]);
            double tmp = Math.abs(direct[i]-indirect[i]);
            if (tmp>error)
                error=tmp;
        }
        
        System.out.println("X");
        printPoints(x);
        System.out.println("Y");
        printPoints(y);
        System.out.println("U");
        printU(u);
        System.out.println("END");
        System.out.println("V");
        printU(indirect);;
        System.out.println("Vdirect");
        printU(direct);
        
       // printTree(t);
        System.out.println("Error: "+error);
        
        System.out.println("Adaptive: "+t.getNumTrans() +" translations, "+t.getNumExp()+" expansions and "+t.getNumDirect()+" direct evaluations with level "+t.getNumOfLevels());
        System.out.println("Nonadaptive: "+tt.getNumTrans() + " (re)expansions with level "+tt.getNumOfLevels());
         
    }
    
    private static void printTree(FmmTree t){
        //System.out.println(t.toString());
        t.print();
    }
    private static void loadInput(Vector<Point> x, Vector<Point> y, double[] u) throws Exception{
        String filename = "D:\\My Documents\\My Programming Projects\\fmmdemo-adaptive2\\input\\input.txt";
        java.io.BufferedReader fr = new java.io.BufferedReader(new java.io.FileReader(filename));
        String line = fr.readLine();//skip the "X"
        line = fr.readLine();
        while (line!=null && !line.equals("Y")){
            Point p = parsePoint(line);
            x.addElement(p);
            line = fr.readLine();
        }
        line=fr.readLine();//skip the "Y"
        while (line!=null&&!line.equals("U")){
            Point p = parsePoint(line);
            y.addElement(p);
            line = fr.readLine();
        }
        line = fr.readLine();
        int count = 0;
        while (line!=null&&!line.equals("END")){
            u[count++]=Double.parseDouble(line);
            line = fr.readLine();
        }
    }
    private static Point parsePoint(String line){
        StringTokenizer ss = new StringTokenizer(line, " ");
        Point p = new Point(Double.parseDouble(ss.nextToken()), Double.parseDouble(ss.nextToken()));
        return p;
    }
    private static void printPoints(Point[] x){
        for (Point p:x){
            System.out.println(p.getCoord().real()+" "+p.getCoord().imaginary());
        }
    }
    private static void printPoints(Vector<Point> x){
        for (Point p:x){
            System.out.println(p.getCoord().real()+" "+p.getCoord().imaginary());
        }
    }
    private static void printU(double[] u){
        for (double d:u)
            System.out.println(d);
    }
    
}
