/*
 * TarWriter.java
 *
 * Created on October 5, 2004, 9:24 AM
 */

package edu.umiacs.fmm.gui;
import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.Color;
import edu.umiacs.fmm.*;
/**
 *
 * @author  wpwy
 */
public class AnimationWorker {
    private int status = Constants.ANIM_STATUS_STOPPED;
    private int command = Constants.ANIM_COMMAND_NOCOMMAND;
    private long interval = Constants.ANIM_INTERVAL_PLAY;
    private LayerAnimation anim = null;
    
    
    /**
     * Utility field used by bound properties.
     */
    private java.beans.PropertyChangeSupport propertyChangeSupport =  new java.beans.PropertyChangeSupport(this);
    
    /**
     * Holds value of property statusMessage.
     */
    private String statusMessage;
    
    /** Creates a new instance of TarWriter */
    public AnimationWorker(LayerAnimation anim) {
        this.anim = anim;
    }
    
    public long getInterval() {
        return interval;
    }
    public void setInterval(long interval){
        this.interval= interval;
    }
    public int getCommand(){
        return command;
    }
    public void setCommand(int c){
        this.command=c;
    }
    
    public void go() {
        final SwingWorker worker = new SwingWorker() {
            public Object construct() {
                setStatus(Constants.ANIM_STATUS_STOPPED);
                setStatusMessage("");
                return new ActualTask();
            }
        };
        worker.start();
    }
    
    public void stop() {
        setCommand(Constants.ANIM_COMMAND_STOP);
        //setStatus(Constants.ANIM_STATUS_STOPPED);
        //setStatusMessage("");
    }
    
    private void processCommandandRepaint() throws AnimationException {
        if (command==Constants.ANIM_COMMAND_PAUSE){
            interval = Constants.ANIM_INTERVAL_PLAY;
            while (command==Constants.ANIM_COMMAND_PAUSE){
                sleepInterval();
            }
        }
        if (command==Constants.ANIM_COMMAND_FF){
            interval = Constants.ANIM_INTERVAL_FF4X;
            sleepInterval();
        } else if (command==Constants.ANIM_COMMAND_FF2END){
            interval = 0;
            //sleepInterval();
        } else if (command==Constants.ANIM_COMMAND_STOP || command==Constants.ANIM_COMMAND_RESET){
            throw new AnimationException();
        }
        if (interval!=0)
            anim.repaint();
    }
    
    //sleeps for a period of time indicated by the current value of "interval"
    private void sleepInterval(){
        sleepInterval(interval);
    }
    //sleeps for a period of time indicated by the i
    private void sleepInterval(long i){
        if (i>0) {
            try {
                Thread.sleep(i);
            } catch(Exception e){
            }
        }
    }
    private void nextStep(){
        //anim.repaint();
        anim.resetHighlightBoxes();
        anim.resetLines();
        if (interval==0) { //FF2END called in this last step
            interval = Constants.ANIM_INTERVAL_PLAY;
            command= Constants.ANIM_COMMAND_PAUSE;
        }
        sleepInterval();
        if (interval!=0)
            anim.repaint();
    }
    
    
    private void animateLines(Vector<edu.umiacs.fmm.Point> from, Vector<edu.umiacs.fmm.Point> to, Color c, int el) throws AnimationException{
        anim.resetLines();
        int numOfLevels = anim.getFmmTree().getStruct().size();
        int lineWidth = numOfLevels-el;
        if (from==null||to==null ||from.size()==0||to.size()==0)
            return;
        for (int k=0; k<=Constants.ANIM_ARROW_STEPS; k++){
            
            anim.resetLines();
            for (edu.umiacs.fmm.Point fromPoint:from){
                for (edu.umiacs.fmm.Point toPoint:to) {
                    double fromX = fromPoint.getCoord().real();
                    double fromY = fromPoint.getCoord().imaginary();
                    double toX = toPoint.getCoord().real();
                    double toY = toPoint.getCoord().imaginary();
                    //int fromXint = (int)(fromX*anim.getProportion());
                    //int fromYint = (int)(fromY*anim.getProportion());
                    //int toXint = (int)((fromX+ (toX-fromX)*k/Constants.ANIM_ARROW_STEPS)*anim.getProportion());
                    //int toYint = (int)((fromY+ (toY-fromY)*k/Constants.ANIM_ARROW_STEPS)*anim.getProportion());
                    double toXIntermediate = fromX+ (toX-fromX)*k/Constants.ANIM_ARROW_STEPS;
                    double toYIntermediate = fromY+ (toY-fromY)*k/Constants.ANIM_ARROW_STEPS;
                    anim.addLine(new AnimLine(fromX, fromY, toXIntermediate, toYIntermediate, c, lineWidth));
                }
            }
            
            sleepInterval();
            processCommandandRepaint();
        }
        //sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
        //anim.resetLines();
    }
    private void animateLines(Vector from, edu.umiacs.fmm.Box to, Color c, int el)throws AnimationException{
        if (from!=null&&from.size()>0) {
            if (from.elementAt(0).getClass().toString().indexOf("Point")>0)
                animateLinesPoint(from, to, c, el);
            else animateLinesBox(from, to, c, el);
        }
    }
    private void animateLines(Vector from, edu.umiacs.fmm.Point to, Color c, int el)throws AnimationException{
        if (from!=null&&from.size()>0) {
            if (from.elementAt(0).getClass().toString().indexOf("Point")>0)
                animateLinesPoint(from, to, c, el);
            else animateLinesBox(from, to, c, el);
        }
    }
    private void animateLinesPoint(Vector<edu.umiacs.fmm.Point> from, edu.umiacs.fmm.Box to, Color c, int el)throws AnimationException{
        Vector<edu.umiacs.fmm.Point> toArr = new Vector<edu.umiacs.fmm.Point>(1);
        toArr.addElement(to.getCenter());
        animateLines(from, toArr,c, el);
    }
    private void animateLinesPoint(Vector<edu.umiacs.fmm.Point> from, edu.umiacs.fmm.Point to, Color c, int el)throws AnimationException{
        Vector<edu.umiacs.fmm.Point> toArr = new Vector<edu.umiacs.fmm.Point>(1);
        toArr.addElement(to);
        animateLines(from, toArr,c, el);
    }
    private void animateLinesBox(Vector<edu.umiacs.fmm.Box> from, edu.umiacs.fmm.Box to, Color c, int el)throws AnimationException{
        Vector<edu.umiacs.fmm.Point> toArr = new Vector<edu.umiacs.fmm.Point>(1);
        toArr.addElement(to.getCenter());
        Vector<edu.umiacs.fmm.Point> fromArr= new Vector<edu.umiacs.fmm.Point>(from.size());
        for (Box fromB:from)
            fromArr.addElement(fromB.getCenter());
        animateLines(fromArr, toArr, c, el);
    }
    private void animateLinesBox(Vector<edu.umiacs.fmm.Box> from, edu.umiacs.fmm.Point to, Color c, int el)throws AnimationException{
        Vector<edu.umiacs.fmm.Point> toArr = new Vector<edu.umiacs.fmm.Point>(1);
        toArr.addElement(to);
        Vector<edu.umiacs.fmm.Point> fromArr= new Vector<edu.umiacs.fmm.Point>(from.size());
        for (Box fromB:from)
            fromArr.addElement(fromB.getCenter());
        animateLines(fromArr, toArr, c, el);
    }
    private void animateLines(edu.umiacs.fmm.Box from, Vector to, Color c, int el) throws AnimationException{
        if (to!=null&&to.size()>0) {
            if (to.elementAt(0).getClass().toString().indexOf("Point")>0)
                animateLinesPoint(from, to, c, el);
            else animateLinesBox(from, to, c, el);
        }
    }
    private void animateLinesPoint(edu.umiacs.fmm.Box from, Vector<edu.umiacs.fmm.Point> to, Color c, int el)throws AnimationException {
        Vector<edu.umiacs.fmm.Point> fromArr = new Vector<edu.umiacs.fmm.Point>(1);
        fromArr.addElement(from.getCenter());
        animateLines(fromArr, to,c, el);
    }
    private void animateLinesBox(edu.umiacs.fmm.Box from, Vector<edu.umiacs.fmm.Box> to, Color c, int el)throws AnimationException{
        Vector<edu.umiacs.fmm.Point> fromArr = new Vector<edu.umiacs.fmm.Point>(1);
        fromArr.addElement(from.getCenter());
        Vector<edu.umiacs.fmm.Point> toArr= new Vector<edu.umiacs.fmm.Point>(to.size());
        for(Box toB:to)
            toArr.addElement(toB.getCenter());
        animateLines(fromArr, toArr, c, el);
    }
    
    private void addCircle(edu.umiacs.fmm.Box b, Color c, int el){
        
        int numOfLevels = anim.getFmmTree().getStruct().size();
        int r = numOfLevels-el;
        int rr = 0;
        if (c.equals(Constants.ANIM_COLOR_S)) {
            r+=4;
            rr=r-2;
        } else if (c.equals(Constants.ANIM_COLOR_SS)) {
            r+=4;
            rr=r-2;
        } else if (c.equals(Constants.ANIM_COLOR_SR)) {
            rr=0;
        } else if (c.equals(Constants.ANIM_COLOR_RR)) {
            r+=2;
            rr=r-2;
        }
        Complex center = b.getCenter().getCoord();
        anim.addCircle(new AnimCircle(center.real(),center.imaginary(), r,rr, c));
    }
    private void addCircles(Vector<edu.umiacs.fmm.Box> bv, Color c, int el){
        for (edu.umiacs.fmm.Box b:bv)
            addCircle(b, c, el);
    }
    
    private void runAnimationUpwardPass() throws AnimationException{
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing upward pass.\n");
        
        java.util.Vector<java.util.Vector<Box>> struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.size();
        for (int l=struct.size()-1; l>=2; l--){
            setStatusMessage("Performing upward pass level "+l+".\n");
            for (Box b:struct.elementAt(l)){
                if (!b.isDummy() && b.getXRecursive().size()>0){
                    setStatusMessage("          "+b.toString()+"\n");
                    if (!b.hasChildren()){ //black node
                        anim.resetHighlightBoxes();
                        anim.addHighlightBox(b);
                        processCommandandRepaint();
                        sleepInterval();
                        
                        animateLines(b.getX(), b, Constants.ANIM_COLOR_S, numOfLevels-1);
                        addCircle(b, Constants.ANIM_COLOR_SS, numOfLevels-1);
                        processCommandandRepaint();
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        anim.resetLines();
                    } else {//gray node
                        //System.out.println(thisBox.toString());
                        anim.resetHighlightBoxes();
                        anim.addHighlightBox(b);
                        Vector<Point> fromVec = new Vector<Point>();
                        for (Box childB:b.getChildren()) {
                            if (childB.getXRecursive().size()>0)
                                fromVec.addElement(childB.getCenter());
                        }
                        animateLines(fromVec, b, Constants.ANIM_COLOR_S, numOfLevels-1);
                        addCircle(b, Constants.ANIM_COLOR_SS, numOfLevels-1);
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        processCommandandRepaint();
                        anim.resetLines();
                    }
                    //else white node.  do nothing.
                }//else a dummy node.  ignore.
            }
        }
    }
    
    private void runAnimationDownwardPass()throws AnimationException{
        processCommandandRepaint();
        //downward pass recursive
        Vector<Vector<edu.umiacs.fmm.Box>> struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.size();
        
        for (int l = 2; l<numOfLevels; l++) {
            
            setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
            setStatusMessage("Performing downward pass.\n");
            setStatusMessage(".....Level: "+l+"\n");
            for (Box b:struct.elementAt(l)){
                if (!b.isDummy() && b.getYRecursive().size()>0){
                    setStatusMessage("          "+b.toString()+"\n");
                    //l2
                    if (b.getL2().size()>0){
                        anim.resetHighlightBoxes();
                        anim.addHighlightBox(b);
                        anim.resetLines();
                        setStatusMessage("              SR translation in L2\n");
                        anim.addHighlightBoxes(b.getL2());
                        processCommandandRepaint();
                        Vector<edu.umiacs.fmm.Box> l2NonEmpty = new Vector<edu.umiacs.fmm.Box>();
                        for (edu.umiacs.fmm.Box l2b:b.getL2()){
                            if (l2b.getYRecursive().size()>0)
                                l2NonEmpty.addElement(l2b);
                        }
                        animateLines(l2NonEmpty, b, Constants.ANIM_COLOR_SR, l);
                        addCircle(b, Constants.ANIM_COLOR_SR, l);
                        processCommandandRepaint();
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                        
                    }
                    //L4
                    if (b.getL4().size()>0){
                        anim.resetHighlightBoxes();
                        anim.addHighlightBox(b);
                        anim.resetLines();
                        setStatusMessage("              R expansion in L4\n");
                        anim.resetLines();
                        anim.addHighlightBoxes(b.getL4());
                        processCommandandRepaint();
                        Vector<Point> xVec = new Vector<Point>();
                        for (Box l4b:b.getL4())
                            xVec.addAll(l4b.getXRecursive());
                        animateLines(xVec, b, Constants.ANIM_COLOR_R, l);
                        addCircle(b, Constants.ANIM_COLOR_R, l);
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                    }
                    //RR
                    if (b.hasChildren()){
                        anim.resetHighlightBoxes();
                        anim.addHighlightBox(b);
                        anim.resetLines();
                        setStatusMessage("              RR translation to children\n");
                        anim.resetLines();
                        anim.addHighlightBoxes(b.getChildren());
                        processCommandandRepaint();
                        Vector<edu.umiacs.fmm.Box> nonEmptyChildren = new Vector<edu.umiacs.fmm.Box>();
                        for (edu.umiacs.fmm.Box child:b.getChildren())
                            if (child.getYRecursive().size()>0)
                                nonEmptyChildren.addElement(child);
                        animateLines(b, nonEmptyChildren, Constants.ANIM_COLOR_RR, l);
                        addCircles(b.getChildren(), Constants.ANIM_COLOR_RR, l);
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                    }
                }
            }
            nextStep();
            processCommandandRepaint();
        }
        
    }
    
    public void runAnimationFinalSummation() throws AnimationException {
        
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        processCommandandRepaint();
        setStatusMessage("Performing final summation.\n");
        
        Vector<Vector<edu.umiacs.fmm.Box>> struct = anim.getFmmTree().getStruct();
        int numOfLevels = struct.size();
        
        for (int i=struct.size()-1; i>0; i--){
            for (Box b:struct.elementAt(i)){
                if (!b.isDummy()){
                    if (!b.hasChildren()&& b.getY().size()>0){ //leaf
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                        
                        setStatusMessage("          "+b.toString()+"\n");
                        
                        if (b.getY().size()>0){
                            setStatusMessage("              R expansion for all y\n");
                            anim.addHighlightBox(b);
                            processCommandandRepaint();
                            animateLines(b, b.getY(), Constants.ANIM_COLOR_R, i);
                            sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                            anim.resetHighlightBoxes();
                            anim.resetLines();
                        }
                        
                        
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                        processCommandandRepaint();
                        
                        setStatusMessage("              Direct summation in L1\n");
                        anim.addHighlightBoxes(b.getL1());
                        anim.addHighlightBox(b);
                        processCommandandRepaint();
                        sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                        
                        anim.resetHighlightBoxes();
                        anim.resetLines();
                        processCommandandRepaint();
                        
                        if (b.getL3().size()>0){
                            setStatusMessage("              S expansion in L3\n");
                            anim.addHighlightBoxes(b.getL3());
                            anim.addHighlightBox(b);
                            processCommandandRepaint();
                            Vector<edu.umiacs.fmm.Box> nonEmptyL3 = new Vector<edu.umiacs.fmm.Box>();
                            for (edu.umiacs.fmm.Box l3b:b.getL3())
                                if (l3b.getXRecursive().size()>0)
                                    nonEmptyL3.addElement(l3b);
                            for (Point py:b.getY()){
                                animateLines(nonEmptyL3, py, Constants.ANIM_COLOR_S, i);
                                sleepInterval(interval*Constants.ANIM_ARROW_STEPS*2);
                                //anim.resetHighlightBoxes();
                                anim.resetLines();
                            }
                        }
                        processCommandandRepaint();
                    }
                }
            }
            nextStep();
        }
    }
    public void runAnimation() throws AnimationException{
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        setStatusMessage("FMM algorithm started.\n");
        setStatus(Constants.ANIM_STATUS_RUNNING);
        sleepInterval();
        runAnimationUpwardPass();
        propertyChangeSupport.firePropertyChange("pass", "up", "down");
        runAnimationDownwardPass();
        runAnimationFinalSummation();
        setStatusMessage(Constants.ANIM_STATUS_MSG_SEPARATOR);
        setStatusMessage("FMM algorithm completed\n");
        setStatus(Constants.ANIM_STATUS_COMPLETED);
    }
    
    /**
     * Adds a PropertyChangeListener to the listener list.
     * @param l The listener to add.
     */
    public void addPropertyChangeListener(java.beans.PropertyChangeListener l) {
        
        propertyChangeSupport.addPropertyChangeListener(l);
    }
    
    /**
     * Removes a PropertyChangeListener from the listener list.
     * @param l The listener to remove.
     */
    public void removePropertyChangeListener(java.beans.PropertyChangeListener l) {
        
        propertyChangeSupport.removePropertyChangeListener(l);
    }
    
    /**
     * Getter for property statusMessage.
     * @return Value of property statusMessage.
     */
    public String getStatusMessage() {
        
        return this.statusMessage;
    }
    
    /**
     * Setter for property statusMessage.
     * @param statusMessage New value of property statusMessage.
     */
    public void setStatusMessage(String statusMessage) {
        
        String oldStatusMessage = this.statusMessage;
        this.statusMessage = statusMessage;
        //System.out.println("AnimationWorker: setting status message to "+statusMessage);
        propertyChangeSupport.firePropertyChange("statusMessage", oldStatusMessage, statusMessage);
    }
    
    /**
     * Getter for property status.
     * @return Value of property status.
     */
    public int getStatus() {
        
        return this.status;
    }
    
    /**
     * Setter for property status.
     * @param status New value of property status.
     */
    public void setStatus(int status) {
        
        int oldStatus = this.status;
        this.status = status;
        //System.out.println("AnimationWorker: setting status to "+status);
        propertyChangeSupport.firePropertyChange("status", new Integer(oldStatus), new Integer(status));
    }
    
    /**
     * The actual long running task.  This runs in a SwingWorker thread.
     */
    class ActualTask {
        ActualTask() {
            try{
                runAnimation();
            } catch (AnimationException e){
                //System.out.println("AnimationException CAUGHT");
                setStatus(Constants.ANIM_STATUS_STOPPED);
                setCommand(Constants.ANIM_COMMAND_NOCOMMAND);
            }
        }
        
    }
}
